"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var math_expression = require("math-expression-evaluator");
var _cloneDeep = require("lodash/cloneDeep");
var _isEqual = require("lodash/isEqual");
var _isNaN = require("lodash/isNaN");
var _isNumber = require("lodash/isNumber");
var _some = require("lodash/some");
var _toNumber = require("lodash/toNumber");
var miller_rabin_1 = require("../algorithm/miller_rabin");
function get_domain_functions(runtime_manager, runtime_data, event_bus, ohno, day_names, config) {
    var TIME_FUNCTIONS = {
        year: function () {
            return new Date().getFullYear();
        },
        month: function () {
            return new Date().getMonth() + 1;
        },
        date: function () {
            return new Date().getDate();
        },
        week: function () {
            return day_names[(new Date().getDay())];
        },
        hour: function () {
            return new Date().getHours();
        },
        minute: function () {
            return new Date().getMinutes();
        },
        second: function () {
            return new Date().getSeconds();
        },
    };
    var LOGIC_FUNCTIONS = {
        EQ: function (args) {
            return _isEqual(args.A, args.B);
        },
        NEQ: function (args) {
            return !_isEqual(args.A, args.B);
        },
        LT: function (args) {
            return args.A < args.B;
        },
        LTE: function (args) {
            return args.A <= args.B;
        },
        GT: function (args) {
            return args.A > args.B;
        },
        GTE: function (args) {
            return args.A >= args.B;
        },
        AND: function (args) {
            return args.A && args.B;
        },
        OR: function (args) {
            return args.A || args.B;
        },
        TRUE: true,
        FALSE: false,
    };
    function to_fixed_number(num, decimals) {
        var pow = Math.pow(10, decimals);
        return Math.round(num * pow) / pow;
    }
    var MATH_FUNCTIONS = {
        ROUND: function (args) {
            return Math.floor(args.NUM + 0.5);
        },
        ROUNDUP: function (args) {
            return Math.ceil(args.NUM);
        },
        ROUNDDOWN: function (args) {
            return Math.floor(args.NUM);
        },
        EVEN: function (args) {
            return args.NUMBER_TO_CHECK % 2 === 0;
        },
        ODD: function (args) {
            return Math.abs(args.NUMBER_TO_CHECK % 2) === 1;
        },
        PRIME: function (args) {
            return miller_rabin_1.is_prime(args.NUMBER_TO_CHECK);
        },
        WHOLE: function (args) {
            return args.NUMBER_TO_CHECK === Math.floor(args.NUMBER_TO_CHECK);
        },
        POSITIVE: function (args) {
            return args.NUMBER_TO_CHECK > 0;
        },
        NEGATIVE: function (args) {
            return args.NUMBER_TO_CHECK < 0;
        },
        DIVISIBLE_BY: function (args) {
            return args.NUMBER_TO_CHECK % args.DIVISOR === 0;
        },
        SIN: function (args) {
            return to_fixed_number(Math.sin(args.NUM * Math.PI / 180), 2);
        },
        ASIN: function (args) {
            return to_fixed_number(Math.asin(args.NUM) / Math.PI * 180, 2);
        },
        COS: function (args) {
            return to_fixed_number(Math.cos(args.NUM * Math.PI / 180), 2);
        },
        ACOS: function (args) {
            return to_fixed_number(Math.acos(args.NUM) / Math.PI * 180, 2);
        },
        TAN: function (args) {
            return to_fixed_number(Math.tan(args.NUM * Math.PI / 180), 2);
        },
        ATAN: function (args) {
            return to_fixed_number(Math.atan(args.NUM) / Math.PI * 180, 2);
        },
        ROOT: function (args) {
            return Math.pow(args.NUM, 0.5);
        },
        ABS: function (args) {
            return Math.abs(args.NUM);
        },
        NEG: function (args) {
            return -args.NUM;
        },
        LN: function (args) {
            return Math.log(args.NUM);
        },
        LOG10: function (args) {
            return Math.log(args.NUM) / Math.log(10);
        },
        EXP: function (args) {
            return Math.pow(Math.E, args.NUM);
        },
        POW10: function (args) {
            return Math.pow(10, args.NUM);
        },
        ADD: function (args) {
            return args.A + args.B;
        },
        MINUS: function (args) {
            return args.A - args.B;
        },
        MULTIPLY: function (args) {
            return args.A * args.B;
        },
        DIVIDE: function (args) {
            return args.A / args.B;
        },
        POWER: function (args) {
            return Math.pow(args.A, args.B);
        },
    };
    function list_update(block_id, entity_id, list) {
        var list_id = runtime_manager.get_list_id(list, entity_id);
        if (!list_id) {
            return;
        }
        if (runtime_manager.is_entity_variable(list_id)) {
            runtime_data.report_entity_list_updated(list_id, list, entity_id);
            return;
        }
        runtime_data.report_list_updated(list_id, list);
    }
    var fns = {
        text_indexOf: function (args, rbid, entity_id) {
            if (args.END === 'FIRST') {
                return args.VALUE.indexOf(args.FIND) + 1;
            }
            else if (args.END === 'LAST') {
                return args.VALUE.lastIndexOf(args.FIND) + 1;
            }
        },
        text_append: function (args, rbid, entity_id) {
            runtime_manager.set_variable(args.VAR, runtime_manager.get_variable(args.VAR, rbid, entity_id) + args.TEXT, rbid, entity_id);
        },
        text_implicit: function (args, rbid, entity_id) {
            var text = args.TEXT;
            if (text === '') {
                return '';
            }
            return text == Number(text) ? Number(text) : text;
        },
        text_join: function (args, rbid, entity_id) {
            var n_text = Object.keys(args).length - 1;
            var s = '';
            var i = 0;
            while (true) {
                var str = args['ADD' + i];
                if (str === undefined) {
                    break;
                }
                s += str + '';
                i++;
            }
            return s;
        },
        text_length: function (args, rbid, entity_id) {
            return (isNaN(args.VALUE)) ? args.VALUE.length : args.VALUE.toString().length;
        },
        text_select: function (args, rbid, entity_id) {
            var str = (isNaN(args.string)) ? args.string : args.string.toString();
            var start_index = args.char_start_index;
            var end_index = args.char_end_index - 1;
            if (start_index <= end_index) {
                return str.substring(start_index - 1, end_index + 1);
            }
            else {
                var result = str.substring(end_index, start_index);
                return result.split('').reverse().join('');
            }
        },
        text_contain: function (args, rbid, entity_id) {
            var text1 = (isNaN(args.TEXT1)) ? args.TEXT1 : args.TEXT1.toString();
            var text2 = (isNaN(args.TEXT2)) ? args.TEXT2 : args.TEXT2.toString();
            return (text1.indexOf(text2) !== -1);
        },
        logic_operation: function (args, rbid, entity_id) {
            return LOGIC_FUNCTIONS[args.OP](args);
        },
        logic_negate: function (args, rbid, entity_id) {
            return !args.BOOL;
        },
        logic_boolean: function (args, rbid, entity_id) {
            return LOGIC_FUNCTIONS[args.BOOL];
        },
        logic_compare: function (args, rbid, entity_id) {
            return LOGIC_FUNCTIONS[args.OP](args);
        },
        math_arithmetic: function (args, rbid, entity_id) {
            return MATH_FUNCTIONS[args.OP](args);
        },
        math_number_property: function (args, rbid, entity_id) {
            return MATH_FUNCTIONS[args.PROPERTY](args);
        },
        divisible_by: function (args, rbid, entity_id) {
            return MATH_FUNCTIONS['DIVISIBLE_BY'](args);
        },
        math_modulo: function (args, rbid, entity_id) {
            return args.DIVIDEND % args.DIVISOR;
        },
        math_trig: function (args, rbid, entity_id) {
            return MATH_FUNCTIONS[args.OP](args);
        },
        math_single: function (args, rbid, entity_id) {
            return MATH_FUNCTIONS[args.OP](args);
        },
        math_round: function (args, rbid, entity_id) {
            return MATH_FUNCTIONS[args.OP](args);
        },
        start_on_click: function () {
        },
        get_time: function (args, rbid, entity_id) {
            return TIME_FUNCTIONS[args.op](args);
        },
        get_cur_frames: function (args, rbid, entity_id) {
            return runtime_manager.get_elapsed_frames();
        },
        variables_get: function (args, rbid, entity_id) {
            return runtime_manager.get_variable(args.VAR, rbid, entity_id);
        },
        variables_set: function (args, rbid, entity_id) {
            runtime_manager.set_variable(args.VAR, args.VALUE, rbid, entity_id);
        },
        days_after_2000: function () {
            var days = (new Date().valueOf() - new Date('2000-01-01').valueOf()) / (1000 * 60 * 60 * 24);
            return Math.floor(days);
        },
        self_broadcast: function (args, rbid, entity_id) {
            var message = args.message;
            runtime_manager.send_action({
                id: 'broadcast',
                namespace: '',
                parameters: undefined,
                sub_type: message,
                value: 'on',
            });
        },
        self_listen: function (args, rbid, entity_id) {
            var value = runtime_data.get_action_state_value({
                action_id: 'broadcast',
                action_namespace: '',
                sub_type: args.message,
            });
            return value === 'on';
        },
        when: function (args) {
            return args.condition;
        },
        wait: function (args, rbid, entity_id) {
            runtime_manager.thread_wait(entity_id, rbid, args.time * 1000);
        },
        change_variable: function (args, rbid, entity_id) {
            var var_id = args.valname;
            var method = args.method || 'increase';
            var n = args.n;
            var value = runtime_manager.get_variable(var_id, rbid, entity_id);
            if (typeof (value) === 'object') {
                return;
            }
            if (method === 'increase') {
                if (_isNumber(_toNumber(value)) && !_isNaN(_toNumber(value))) {
                    value = _toNumber(value);
                }
                runtime_manager.set_variable(var_id, value + n, rbid, entity_id);
            }
            else {
                runtime_manager.set_variable(var_id, (value - n), rbid, entity_id);
            }
        },
        random: function (args, rbid, entity_id) {
            var a = args.a;
            var b = args.b;
            var rand = runtime_manager.get_random_number();
            var randomFunc = Math.floor(a + (rand * (b - a + 1)));
            return randomFunc;
        },
        reset_timer: function (args, rbid, entity_id) {
            runtime_manager.reset_timer();
        },
        destruct: function (args, interpreter_id, entity_id, internals) {
            internals.runtime_manager.destruct_entity(entity_id);
        },
        get_timer: function (args, rbid, entity_id) {
            return runtime_manager.get_timer_elapsed_s();
        },
        terminate: function () {
            runtime_manager.stop();
        },
        restart: function (args, rbid, entity_id) {
            runtime_manager.restart();
        },
        stop: function (args, rbid, entity_id) {
            var type = parseInt(args.scope);
            if (type === 0) {
                runtime_manager.dispose_all();
            }
            else if (type === 1) {
                runtime_manager.dispose_block_group(rbid);
            }
            else if (type === 2) {
                runtime_manager.dispose_other_block_groups_of_entity(entity_id, rbid);
            }
            else if (type === 3) {
                runtime_manager.dispose_block_groups_of_other_entities(entity_id);
            }
        },
        lists_get: function (args, rbid, entity_id) {
            return runtime_manager.lists_get(args.VAR, rbid, entity_id);
        },
        lists_append: function (args, rbid, entity_id) {
            var list = args.VAR;
            list.push(args.VALUE);
            list_update(rbid, entity_id, list);
        },
        lists_delete: function (args, rbid, entity_id) {
            var index = args.INDEX;
            var list = args.VAR;
            if (index === 0) {
                return;
            }
            if (args.TYPE === 'last') {
                list.pop();
            }
            else {
                index = index > 0 ? index - 1 : index;
                list.splice(index, 1);
            }
            list_update(rbid, entity_id, list);
        },
        lists_insert_value: function (args, rbid, entity_id) {
            var index = args.INDEX;
            var list = args.VAR;
            if (index === 0) {
                return;
            }
            if (index === -1) {
                list.push(args.VALUE);
            }
            else {
                index = index > 0 ? index - 1 : index + 1;
                list.splice(index, 0, args.VALUE);
            }
            list_update(rbid, entity_id, list);
        },
        lists_replace: function (args, rbid, entity_id) {
            var index = args.INDEX;
            var list = args.VAR;
            if (index === 0) {
                return;
            }
            if (args.TYPE === 'last') {
                index = list.length;
            }
            index = index > 0 ? index - 1 : index;
            list.splice(index, 1, args.VALUE);
            list_update(rbid, entity_id, list);
        },
        lists_get_value: function (args, rbid, entity_id) {
            var index = args.INDEX;
            var list = args.VAR;
            if (args.TYPE === 'last') {
                index = list.length;
            }
            if (index === 0) {
                if (config.get().legacy.lists_get_value_allow_return_undefined == false) {
                    throw ohno.user.lists_get_value_bad_index({
                        list: list,
                        index: -1,
                        args: args,
                    });
                }
                return;
            }
            else {
                index = index > 0 ? index - 1 : list.length + index;
                var result = list[index];
                if (result == undefined && config.get().legacy.lists_get_value_allow_return_undefined == false) {
                    throw ohno.user.lists_get_value_bad_index({
                        list: list,
                        index: index > 0 ? index : -1,
                        args: args,
                    });
                }
                return result;
            }
        },
        lists_is_exist: function (args, rbid, entity_id) {
            var stringified = args.VALUE.toString();
            var list = args.VAR;
            return _some(list, function (e) { return e == args.VALUE || e == stringified; });
        },
        lists_index_of: function (args, rbid, entity_id) {
            var text = args.VALUE;
            var list = args.VAR;
            var val = list.indexOf(text);
            return typeof val === 'number' ? val + 1 : 0;
        },
        lists_length: function (args, rbid, entity_id) {
            var list = args.VAR;
            return list.length;
        },
        calculate: function (args, rbid, entity_id) {
            try {
                return math_expression.lex(args.input).toPostfix().postfixEval();
            }
            catch (e) {
                throw ohno.user.block_bad_math_expression({ block_args: args });
            }
        },
        lists_copy: function (args, rbid, entity_id) {
            var value = _cloneDeep(args.VALUE);
            var target_list = args.TARGET;
            target_list.length = 0;
            value = value || [];
            for (var i = 0; i < value.length; i++) {
                var val = value[i];
                target_list.push(val);
            }
            list_update(rbid, entity_id, target_list);
        },
        text_split: function (args, rbid, entity_id) {
            var text_to_split = args.TEXT_TO_SPLIT;
            if (typeof text_to_split === 'number') {
                text_to_split = text_to_split.toString();
            }
            var split_text = text_to_split.split(args.SPLIT_TEXT) || [];
            return split_text.map(function (element) {
                var n = element.indexOf(' ') < 0 && element !== '' ?
                    _toNumber(element) :
                    element;
                return _isNaN(n) ? element : n;
            });
        },
    };
    return function () { return fns; };
}
exports.get_domain_functions = get_domain_functions;
