"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var _cloneDeep = require("lodash/cloneDeep");
var _last = require("lodash/last");
var _sortBy = require("lodash/sortBy");
var _identity = require("lodash/identity");
var _defaultsDeep = require("lodash/defaultsDeep");
var catastrophic_1 = require("catastrophic");
var inversify_1 = require("inversify");
var BLK = require("./block_types");
var di_symbols_1 = require("./di_symbols");
var basic_types_1 = require("./basic_types");
var BLOCK_RES;
(function (BLOCK_RES) {
    BLOCK_RES[BLOCK_RES["procedure_call"] = 1] = "procedure_call";
    BLOCK_RES[BLOCK_RES["procedure_return"] = 2] = "procedure_return";
    BLOCK_RES[BLOCK_RES["run_child"] = 3] = "run_child";
    BLOCK_RES[BLOCK_RES["value"] = 4] = "value";
    BLOCK_RES[BLOCK_RES["empty"] = 5] = "empty";
    BLOCK_RES[BLOCK_RES["null_or_undefined_error"] = 6] = "null_or_undefined_error";
    BLOCK_RES[BLOCK_RES["enter_warp"] = 7] = "enter_warp";
})(BLOCK_RES || (BLOCK_RES = {}));
function is_result_procedure_call(r) {
    return BLOCK_RES[r.kind] == BLOCK_RES.procedure_call;
}
function is_result_procedure_return(r) {
    return BLOCK_RES[r.kind] == BLOCK_RES.procedure_return;
}
function is_result_run_child(r) {
    return BLOCK_RES[r.kind] == BLOCK_RES.run_child;
}
function is_result_empty(r) {
    return BLOCK_RES[r.kind] == BLOCK_RES.empty;
}
function is_null_or_undefined_error(r) {
    return BLOCK_RES[r.kind] == BLOCK_RES.null_or_undefined_error;
}
function is_enter_warp(r) {
    return BLOCK_RES[r.kind] == BLOCK_RES.enter_warp;
}
function is_result_value(r) {
    return BLOCK_RES[r.kind] == BLOCK_RES.value;
}
function res_empty() {
    return {
        kind: 'empty',
    };
}
function res_null_or_undefined_error() {
    return {
        kind: 'null_or_undefined_error',
    };
}
function res_value(val) {
    if (val == undefined) {
        return { kind: 'empty' };
    }
    return {
        kind: 'value',
        value: val,
    };
}
function res_run_child(child) {
    return {
        kind: 'run_child',
        child_block: child,
    };
}
function res_procedure_call(yield_to_block, procedure, args, call_block) {
    return {
        kind: 'procedure_call',
        yield_to_block: yield_to_block,
        procedure_block: procedure,
        call_arguments: args,
        call_block: call_block,
    };
}
function res_procedure_return(return_value) {
    return {
        kind: 'procedure_return',
        return_value: return_value,
    };
}
var BlockInterpreterImpl = (function () {
    function BlockInterpreterImpl(util, ohno, run_mgr, block_pool, block_runner, priorities, compiled_block, script_variable_specs, original_identities, group_id, should_report_current_running_block, max_procedure_calls_per_interpreter_step, max_warp_iterations_per_interpreter_step, max_call_stack_size, is_warped, on_finished) {
        this.util = util;
        this.ohno = ohno;
        this.run_mgr = run_mgr;
        this.block_pool = block_pool;
        this.block_runner = block_runner;
        this.priorities = priorities;
        this.original_identities = original_identities;
        this.max_procedure_calls_per_interpreter_step = max_procedure_calls_per_interpreter_step;
        this.max_warp_iterations_per_interpreter_step = max_warp_iterations_per_interpreter_step;
        this.max_call_stack_size = max_call_stack_size;
        this.is_warped = is_warped;
        this.on_finished = on_finished;
        this.procedure_call_stack = [];
        this.variables = {};
        this.variables_stack = [];
        this.should_report_current_running_block = false;
        this.n_warp_iterations = 0;
        this.n_warp_parents = 0;
        this.metadata = {
            priorities: priorities,
            original_identities: original_identities,
            typeclass_id: original_identities.typeclass_id,
            interpreter_id: original_identities.interpreter_id,
            type: compiled_block.type,
            group_id: group_id,
        };
        this.current_block = compiled_block;
        this.root_block = this.current_block;
        this.root_block_id = compiled_block.id;
        for (var var_id in script_variable_specs) {
            this.variables[var_id] = script_variable_specs[var_id].value;
        }
        this.original_variables = _cloneDeep(this.variables);
        this.should_report_current_running_block = should_report_current_running_block;
    }
    BlockInterpreterImpl.prototype.reset = function () {
        this.current_block = this.block_pool.clone(this.run_mgr.get_compiled_block_by_interpreter_id(this.root_block_id));
        this.block_pool.release(this.root_block);
        this.root_block = this.current_block;
    };
    BlockInterpreterImpl.prototype.dispose = function () {
        this.block_pool.release(this.root_block);
    };
    BlockInterpreterImpl.prototype.is_inside_warp = function () {
        return this.is_warped || this.n_warp_parents > 0;
    };
    BlockInterpreterImpl.prototype.set_variable = function (name, value) {
        this.variables[name] = value;
    };
    BlockInterpreterImpl.prototype.get_variable = function (name) {
        return this.variables[name];
    };
    BlockInterpreterImpl.prototype.get_variables = function () {
        return this.variables;
    };
    BlockInterpreterImpl.prototype.step_into_procedure = function (res) {
        if (res.procedure_block.child_block.length == 0) {
            return false;
        }
        var procedure = res.procedure_block.child_block[0];
        if (procedure == undefined) {
            return false;
        }
        res.yield_to_block.first_evaluation = false;
        res.procedure_block.parent_block = res.yield_to_block;
        this.procedure_call_stack.push(res);
        this.variables_stack.push(this.variables);
        this.variables = _cloneDeep(this.original_variables);
        this.current_block = procedure;
        return true;
    };
    BlockInterpreterImpl.prototype.step_out_of_procedure = function (res) {
        var call = this.procedure_call_stack.pop();
        if (call == undefined) {
            throw this.ohno.system.procedure_popped_empty_call_stack({
                root_block_id: this.root_block_id,
                block: this.current_block,
                block_id: this.current_block != undefined ? this.current_block.id : undefined,
            });
        }
        var proc_stack = this.procedure_call_stack;
        if (proc_stack.length > this.max_call_stack_size) {
            throw this.ohno.system.stack_overflow({
                block_id: this.current_block != undefined ? this.current_block.id : undefined,
                procedure_id: call.procedure_block.id,
            });
        }
        if (call.call_block) {
            call.call_block.procedure_return_value = res.return_value;
        }
        else {
            call.yield_to_block.done_evaluating = true;
        }
        this.current_block = call.yield_to_block;
        var var_stack_frame = this.variables_stack.pop();
        if (var_stack_frame == undefined) {
            throw this.ohno.system.popped_empty_variable_stack({
                root_block_id: this.root_block_id,
                block: this.current_block,
                block_id: this.current_block.id,
            });
        }
        this.variables = var_stack_frame;
    };
    BlockInterpreterImpl.prototype.do_step = function () {
        this.run_mgr.set_current_interpreter_not_blocked();
        this.n_warp_iterations = 0;
        var n_procedure_calls = 0;
        while (this.current_block) {
            if (this.util.is.break(this.current_block)) {
                var first_loop_ancestor = this.util.get_first_ancestor_satisfying(this.current_block, this.util.is.loop_block);
                if (!first_loop_ancestor) {
                    throw this.ohno.user.break_with_bad_parent({
                        root_block_id: this.root_block_id,
                        block: this.current_block,
                        block_id: this.current_block.id,
                    });
                }
                var ancestor = this.current_block.parent_block;
                while (ancestor != first_loop_ancestor && ancestor != undefined) {
                    if (this.util.is.warp(ancestor)) {
                        this.n_warp_parents--;
                    }
                    this.util.reset_state(ancestor);
                    ancestor = ancestor.parent_block;
                }
                first_loop_ancestor.done_evaluating = true;
                this.current_block = first_loop_ancestor;
                continue;
            }
            if (this.current_block.done_evaluating == false) {
                if (this.should_report_current_running_block) {
                    this.run_mgr.set_running_block(this.root_block_id, this.current_block.id);
                }
                var res = this.block_runner.run_block(this.original_identities, this.metadata, this.root_block_id, this.current_block, _last(this.procedure_call_stack));
                if (is_enter_warp(res)) {
                    this.n_warp_parents++;
                    this.current_block = res.child_block;
                    continue;
                }
                if (is_result_run_child(res)) {
                    this.current_block = res.child_block;
                    continue;
                }
                if (is_result_procedure_call(res)) {
                    n_procedure_calls++;
                    if (n_procedure_calls > this.max_procedure_calls_per_interpreter_step) {
                        return basic_types_1.StepResult.yielding;
                    }
                    if (this.step_into_procedure(res)) {
                        continue;
                    }
                }
                if (is_result_procedure_return(res)) {
                    this.step_out_of_procedure(res);
                    continue;
                }
                if (this.util.is.wait_until(this.current_block)) {
                    if (is_result_value(res)) {
                        if (!res.value) {
                            this.util.reset_state(this.current_block);
                            return basic_types_1.StepResult.yielding;
                        }
                    }
                }
            }
            if (this.run_mgr.current_interpreter_must_yield(this.original_identities.interpreter_id, this.metadata.group_id)) {
                this.current_block.done_evaluating = true;
                return basic_types_1.StepResult.yielding;
            }
            this.util.reset_state(this.current_block);
            if (this.current_block.next_block) {
                this.current_block = this.current_block.next_block;
                continue;
            }
            if (this.util.is.event_block(this.current_block)) {
                return basic_types_1.StepResult.yielding;
            }
            if (!this.current_block.parent_block) {
                if (this.on_finished !== undefined) {
                    this.on_finished(this.root_block_id);
                }
                return basic_types_1.StepResult.finished;
            }
            var parent_block = this.current_block.parent_block;
            if (this.util.is.loop_block(parent_block)) {
                if (this.util.is.repeat_forever(parent_block)) {
                    this.current_block = parent_block.child_block[0];
                }
                if (this.util.is.repeat_n_times(parent_block)) {
                    parent_block.times_left--;
                    if (parent_block.times_left > 0) {
                        this.current_block = parent_block.child_block[0];
                    }
                    else {
                        parent_block.done_evaluating = true;
                        this.current_block = parent_block;
                    }
                }
                if (this.util.is.repeat_forever_until(parent_block)) {
                    this.util.reset_state(parent_block);
                    this.current_block = parent_block;
                }
                if (this.is_inside_warp()) {
                    if (this.n_warp_iterations < this.max_warp_iterations_per_interpreter_step) {
                        this.n_warp_iterations++;
                        continue;
                    }
                }
                return basic_types_1.StepResult.yielding;
            }
            if (this.util.is.warp(parent_block)) {
                parent_block.done_evaluating = true;
                this.current_block = parent_block;
                this.n_warp_parents--;
                continue;
            }
            if (this.util.is.cond_block(parent_block)) {
                parent_block.done_evaluating = true;
                this.current_block = parent_block;
                continue;
            }
            if (this.util.is.event_block(parent_block)) {
                if (BLK.EVENT_BLOCKS[parent_block.type] == BLK.EVENT_BLOCKS.start_as_a_mirror) {
                    return basic_types_1.StepResult.finished;
                }
                this.current_block = parent_block;
                return basic_types_1.StepResult.yielding;
            }
            if (this.util.is.procedures_defnoreturn(parent_block)) {
                this.step_out_of_procedure(res_procedure_return(undefined));
                continue;
            }
            if (this.util.is.async_tell(parent_block)) {
                return basic_types_1.StepResult.finished;
            }
            return basic_types_1.StepResult.finished;
        }
        return basic_types_1.StepResult.finished;
    };
    BlockInterpreterImpl.prototype.step = function () {
        try {
            return this.do_step();
        }
        catch (e) {
            var error_metadata = {
                interpreter_id: this.root_block_id,
                interpreter_stack: this.get_current_stack(),
            };
            if (e instanceof catastrophic_1.Catastrophe) {
                e.annotation = _defaultsDeep(e.annotation, error_metadata);
                throw e;
            }
            throw this.ohno.user.unknown_run_block_error(e, error_metadata);
        }
    };
    BlockInterpreterImpl.prototype.get_current_stack = function () {
        var proc_stack = this.procedure_call_stack;
        var stack = [];
        var is_only_frame = proc_stack.length == 0;
        var main_frame = {
            interpreter_id: this.original_identities.interpreter_id,
            source_map_rbid: this.original_identities.source_map_rbid,
            source_entity_id: this.original_identities.source_map_entity,
            block_id: is_only_frame ?
                (this.current_block != undefined) ? this.current_block.id : undefined :
                this.get_procedure_call_block_id(proc_stack[0]),
            proc_id: undefined,
        };
        stack.push(main_frame);
        if (is_only_frame) {
            return stack;
        }
        for (var i = 0; i < proc_stack.length; i++) {
            var is_last_frame = (i == proc_stack.length - 1);
            var proc_id = proc_stack[i].procedure_block.procedure_name;
            stack.push({
                interpreter_id: this.original_identities.interpreter_id,
                source_map_rbid: this.original_identities.source_map_rbid,
                source_entity_id: this.original_identities.source_map_entity,
                block_id: is_last_frame ?
                    (this.current_block != undefined) ? this.current_block.id : undefined :
                    this.get_procedure_call_block_id(proc_stack[i + 1]),
                proc_id: proc_id,
            });
        }
        return stack;
    };
    BlockInterpreterImpl.prototype.get_procedure_call_block_id = function (proc_stack) {
        if (proc_stack.call_block == undefined) {
            return proc_stack.yield_to_block.id;
        }
        return proc_stack.call_block.id;
    };
    return BlockInterpreterImpl;
}());
var BlockRunnerImpl = (function () {
    function BlockRunnerImpl(u, ohno, run_mgr, block_pool, event_bus, domain_functions) {
        var _this = this;
        this.u = u;
        this.ohno = ohno;
        this.run_mgr = run_mgr;
        this.block_pool = block_pool;
        this.event_bus = event_bus;
        this.domain_functions = domain_functions;
        this.deterministic = u.config.get().deterministic != undefined;
        this.tell_should_ensure_entity_exists = u.config.get().reports_all_entities;
        event_bus.system.config_updated.immediate.sub(function () {
            _this.tell_should_ensure_entity_exists = u.config.get().reports_all_entities;
        });
    }
    BlockRunnerImpl.prototype.run_block = function (original_identities, metadata, root_block_id, block, latest_frame) {
        var _this = this;
        if (block == undefined || block == null) {
            return res_null_or_undefined_error();
        }
        if (this.u.block.is.compiled_block(block) && block.disabled) {
            return res_empty();
        }
        try {
            if (this.u.block.is.compiled_block(block) && BLK.ATOMIC_BLOCKS[block.type]) {
                return res_value(block.params[Object.keys(block.params)[0]]);
            }
            if (this.u.block.is.atomic(block)) {
                return res_value(block);
            }
        }
        catch (e) {
            throw this.ohno.user.error_constructing_value_from_atomic_block(e, {
                root_block_id: root_block_id,
                block: block,
            });
        }
        if (block.first_evaluation == true) {
            block.last_call = this.run_mgr.get_elapsed_frames();
        }
        var args = { blockId: root_block_id };
        var procedure_call_to_wait_for;
        var param_ids = Object.keys(block.params);
        if (this.deterministic) {
            param_ids = _sortBy(param_ids, _identity);
        }
        for (var i = 0; i < param_ids.length; i++) {
            var param_id = param_ids[i];
            var param_block = block.params[param_id];
            if (this.u.block.is.compiled_block(param_block) && param_block.disabled) {
                throw this.ohno.compiler.user.disabled_param(new Error(), {
                    root_block_id: root_block_id,
                    block_id: param_block.id,
                    block_type: param_block.type,
                    block: param_block,
                    source_entity_id: original_identities.source_map_entity,
                });
            }
            var res = this.run_block(original_identities, metadata, root_block_id, param_block, latest_frame);
            if (is_result_value(res)) {
                args[param_id] = res.value;
                continue;
            }
            if (is_result_procedure_call(res)) {
                procedure_call_to_wait_for = res;
                break;
            }
            if (is_null_or_undefined_error(res)) {
                throw this.ohno.system.undefined_or_null_block({
                    root_block_id: root_block_id,
                    block: block,
                    block_id: block.id,
                });
            }
            if (is_result_empty(res)) {
                throw this.ohno.user.undefined_code_path_argument({
                    root_block_id: root_block_id,
                    block: block,
                    block_id: block.id,
                });
            }
            throw this.ohno.system.unhandled_run_block_result({
                root_block_id: root_block_id,
                block: block,
                block_id: block.id,
            });
        }
        if (procedure_call_to_wait_for != undefined) {
            return procedure_call_to_wait_for;
        }
        try {
            if (this.u.block.is.cond_block(block)) {
                return this.run_conditional(original_identities, metadata, root_block_id, block, latest_frame);
            }
            else if (this.u.block.is.loop_block(block)) {
                return this.run_loop(root_block_id, block, args);
            }
            else if (this.u.block.is.event_block(block)) {
                return this.run_event(root_block_id, block, args);
            }
            else if (this.u.block.is.responder_block(block)) {
                return res_run_child(block.child_block[0]);
            }
            else if (this.u.block.is.proc_block(block)) {
                return this.run_procedure(root_block_id, block, args, latest_frame);
            }
            else if (this.u.block.is.warp(block)) {
                return {
                    kind: 'enter_warp',
                    child_block: block.child_block[0],
                };
            }
            else if (this.u.block.is.async_tell(block) || this.u.block.is.sync_tell(block)) {
                return this.run_tell(original_identities, metadata, root_block_id, block, args);
            }
            else if (this.u.block.is.logic_empty(block)) {
                return res_value(false);
            }
            if (this.domain_functions[block.type] == undefined) {
                throw this.ohno.system.missing_domain_function({
                    root_block_id: root_block_id,
                    block: block,
                });
            }
            else {
                this.u.block.reset_state(block);
                var domain_fun_result = void 0;
                try {
                    domain_fun_result = (this.domain_functions[block.type])(args, original_identities.interpreter_id, original_identities.target_entity, {
                        runtime_manager: this.run_mgr,
                        add_user_procedure_call_to_stack: function () {
                            throw _this.ohno.system.feature_not_available_in_debug_mode({
                                feature: 'Calling user procedures',
                            });
                        },
                        get_action_parameter: function (parameter_id) {
                            throw _this.ohno.system.feature_not_available_in_debug_mode({
                                feature: 'Getting action parameters',
                            });
                        },
                    });
                }
                catch (e) {
                    if (e instanceof catastrophic_1.Catastrophe) {
                        throw e;
                    }
                    throw this.ohno.system.unknown_error_in_domain_function_call(e);
                }
                if (domain_fun_result != undefined && typeof domain_fun_result.then == 'function') {
                    throw this.ohno.system.feature_not_available_in_debug_mode({
                        feature: 'Async domain functions',
                    });
                }
                var block_result = res_value(domain_fun_result);
                if (domain_fun_result == undefined
                    && block.type == 'lists_get_value'
                    && this.u.config.get().legacy.lists_get_value_allow_return_undefined) {
                    block_result = { kind: 'value', value: undefined };
                }
                if (!is_result_empty(block_result) && block_result.value !== undefined) {
                    if (!block.parent_block) {
                        this.event_bus.runtime_data.block_run_result.send({
                            root_block_id: root_block_id,
                            block_id: block.id,
                            result: block_result.value,
                        });
                    }
                }
                return block_result;
            }
        }
        catch (e) {
            var error_metadata = {
                block: block,
                block_id: block.id,
                block_type: block.type,
                source_entity_id: original_identities.source_map_entity,
            };
            if (e instanceof catastrophic_1.Catastrophe) {
                e.annotation = _defaultsDeep(e.annotation, error_metadata);
                throw e;
            }
            throw this.ohno.user.unknown_run_block_error(e, error_metadata);
        }
    };
    BlockRunnerImpl.prototype.run_conditional = function (original_identities, metadata, root_block_id, block, latest_frame) {
        var i = 0;
        for (; i < block.child_block.length - 1; ++i) {
            var condition = block.conditions[i];
            if (condition == undefined) {
                continue;
            }
            if (condition.disabled) {
                throw this.ohno.compiler.user.disabled_param(new Error(), {
                    root_block_id: root_block_id,
                    block_id: condition.id,
                    block_type: condition.type,
                    block: condition,
                });
            }
            var res = this.run_block(original_identities, metadata, root_block_id, condition, latest_frame);
            if (is_result_procedure_call(res)) {
                return res;
            }
            if (is_result_value(res) && res.value) {
                var then = block.child_block[i];
                if (then != undefined) {
                    return res_run_child(then);
                }
                return res_empty();
            }
        }
        var _else = block.child_block[i];
        if (_else != undefined) {
            return res_run_child(_else);
        }
        return res_empty();
    };
    BlockRunnerImpl.prototype.run_loop = function (root_block_id, block, args) {
        if (this.u.block.is.repeat_forever(block)
            && block.first_evaluation == true) {
            block.first_evaluation = false;
            return res_run_child(block.child_block[0]);
        }
        if (this.u.block.is.repeat_n_times(block)) {
            block.first_evaluation = false;
            block.times_left = args['times'];
            if (block.times_left > 0) {
                return res_run_child(block.child_block[0]);
            }
            block.done_evaluating = true;
            return res_empty();
        }
        if (this.u.block.is.repeat_forever_until(block)) {
            if (args['condition']) {
                block.done_evaluating = true;
                return res_empty();
            }
            return res_run_child(block.child_block[0]);
        }
        if (this.u.block.is.wait_until(block)) {
            return res_value(args['condition']);
        }
        return res_empty();
    };
    BlockRunnerImpl.prototype.find_yielding_ancestor = function (block) {
        var yielding_ancestor = block.parent_block;
        while (yielding_ancestor != undefined && yielding_ancestor.output_type != BLK.BlockOutputType.none) {
            yielding_ancestor = yielding_ancestor.parent_block;
        }
        return yielding_ancestor;
    };
    BlockRunnerImpl.prototype.run_procedure = function (root_block_id, block, args, latest_frame) {
        if (this.u.block.is.procedures_callnoreturn(block)) {
            var procedure = this.run_mgr.get_procedure(block.procedure_name);
            if (procedure == undefined) {
                throw this.u.ohno.user.call_undefined_procedure({
                    block_id: block.id,
                });
            }
            return res_procedure_call(block, procedure.script, args);
        }
        if (this.u.block.is.procedures_callreturn(block)) {
            var yielding_ancestor = this.find_yielding_ancestor(block) || block;
            if (block.procedure_return_value != undefined) {
                if (block.parent_block == undefined) {
                    this.event_bus.runtime_data.block_run_result.send({
                        root_block_id: root_block_id,
                        block_id: block.id,
                        result: block.procedure_return_value,
                    });
                }
                var ancestor_call_time = yielding_ancestor.last_call;
                var current_call_time = block.last_call;
                if (ancestor_call_time == undefined || current_call_time == undefined) {
                    throw this.ohno.system.procedure_missing_call_timestamps({
                        root_block_id: root_block_id,
                        args: args,
                        block: block,
                    });
                }
                if (ancestor_call_time <= current_call_time) {
                    return res_value(block.procedure_return_value);
                }
                block.last_call = this.run_mgr.get_elapsed_frames();
            }
            else {
                block.first_evaluation = false;
            }
            var procedure = this.run_mgr.get_procedure(block.procedure_name);
            if (procedure == undefined) {
                throw this.u.ohno.user.call_undefined_procedure({
                    block_id: block.id,
                });
            }
            return res_procedure_call(yielding_ancestor, procedure.script, args, block);
        }
        if (this.u.block.is.procedures_return_value(block)) {
            if (args['VALUE'] == undefined) {
                throw this.ohno.user.procedure_return_empty({
                    root_block_id: root_block_id,
                    args: args,
                    block: block,
                });
            }
            if (latest_frame == undefined) {
                throw this.ohno.user.procedure_return_outside({
                    root_block_id: root_block_id,
                    args: args,
                    block: block,
                });
            }
            return res_procedure_return(args['VALUE']);
        }
        if (this.u.block.is.procedures_parameter(block)) {
            if (latest_frame == undefined) {
                throw this.ohno.user.procedure_parameter_outside({
                    root_block_id: root_block_id,
                    args: args,
                    block: block,
                });
            }
            var param_name = args['param_name'];
            if (latest_frame.call_arguments[param_name] == undefined) {
                throw this.ohno.user.procedure_no_such_parameter({
                    args: args,
                    param_name: param_name,
                    root_block_id: root_block_id,
                    block: block,
                });
            }
            return res_value(latest_frame.call_arguments[param_name]);
        }
        return res_empty();
    };
    BlockRunnerImpl.prototype.run_event = function (root_block_id, block, args) {
        var _this = this;
        var check = this.domain_functions[block.type](args, root_block_id, this.run_mgr.get_entity_id_from_root_block_id(root_block_id), {
            runtime_manager: this.run_mgr,
            add_user_procedure_call_to_stack: function () {
                throw _this.ohno.system.feature_not_available_in_debug_mode({
                    feature: 'Calling user procedures',
                });
            },
            get_action_parameter: function (parameter_id) {
                throw _this.ohno.system.feature_not_available_in_debug_mode({
                    feature: 'Getting action parameters',
                });
            },
        });
        if (check) {
            return res_run_child(block.child_block[0]);
        }
        return res_empty();
    };
    BlockRunnerImpl.prototype.run_tell = function (original_identities, metadata, root_block_id, block, args) {
        block.first_evaluation = false;
        var child = block.child_block[0];
        if (child == undefined) {
            return res_empty();
        }
        var new_target_entity_id = args.sprite;
        var target_entity_state = this.run_mgr.get_entity_state(new_target_entity_id);
        if (this.tell_should_ensure_entity_exists) {
            if (target_entity_state == basic_types_1.EntityState.Destructing) {
                this.run_mgr.report_warning(this.ohno.warning.tell_with_destructing_entity({
                    args: args,
                    block: block,
                    root_block_id: root_block_id,
                    tell_entity: new_target_entity_id,
                }));
                return res_empty();
            }
            else if (target_entity_state == basic_types_1.EntityState.Disposed) {
                this.run_mgr.report_warning(this.ohno.warning.tell_with_disposed_entity({
                    args: args,
                    block: block,
                    root_block_id: root_block_id,
                    tell_entity: new_target_entity_id,
                }));
                return res_empty();
            }
            else if (target_entity_state == basic_types_1.EntityState.Unknown) {
                throw this.ohno.user.tell_with_unknown_entity({
                    args: args,
                    block: block,
                    root_block_id: root_block_id,
                    tell_entity: new_target_entity_id,
                });
            }
        }
        var identities = {
            typeclass_id: original_identities.typeclass_id,
            interpreter_id: original_identities.interpreter_id,
            target_entity: original_identities.target_entity,
            source_map_entity: original_identities.source_map_entity,
            source_map_rbid: root_block_id,
        };
        if (this.u.block.is.sync_tell(block)) {
            this.run_mgr.spawn_sync_tell_interpreter(identities, new_target_entity_id, this.block_pool.clone(child), metadata.group_id);
            return res_empty();
        }
        this.run_mgr.spawn_async_tell_interpreter(identities, new_target_entity_id, this.block_pool.clone(child), metadata.group_id);
        return res_empty();
    };
    return BlockRunnerImpl;
}());
var BlockInterpreterFactoryImpl = (function () {
    function BlockInterpreterFactoryImpl(u, ohno, event_bus, block_registry, opti_compiler, block_pool) {
        var _this = this;
        this.u = u;
        this.ohno = ohno;
        this.event_bus = event_bus;
        this.block_registry = block_registry;
        this.opti_compiler = opti_compiler;
        this.block_pool = block_pool;
        var configure = function () {
            var conf = u.config.get();
            _this.should_report_current_running_block = conf.should_report_current_running_block;
            _this.max_procedure_calls_per_interpreter_step = conf.max_procedure_calls_per_interpreter_step;
            _this.max_warp_iterations_per_interpreter_step = conf.max_warp_iterations_per_interpreter_step;
            _this.max_call_stack_size = conf.max_call_stack_size;
        };
        configure();
        this.event_bus.system.config_updated.immediate.sub(configure);
    }
    BlockInterpreterFactoryImpl.prototype.create = function (run_mgr, identities, priorities, compile_cache_id, compiled_block, script_variable_specs, group_id, is_warped, action_parameters, on_finished) {
        if (this.block_runner == undefined) {
            this.block_runner = new BlockRunnerImpl(this.u, this.u.ohno, run_mgr, this.block_pool, this.event_bus, this.block_registry.get_domain_functions());
        }
        return new BlockInterpreterImpl(this.u.block, this.ohno, run_mgr, this.block_pool, this.block_runner, priorities, compiled_block, script_variable_specs, identities, group_id, this.should_report_current_running_block, this.max_procedure_calls_per_interpreter_step, this.max_warp_iterations_per_interpreter_step, this.max_call_stack_size, is_warped, on_finished);
    };
    BlockInterpreterFactoryImpl.prototype.clear = function () { };
    BlockInterpreterFactoryImpl = tslib_1.__decorate([
        inversify_1.injectable(),
        tslib_1.__param(0, inversify_1.inject(di_symbols_1.BINDING.Util)),
        tslib_1.__param(1, inversify_1.inject(di_symbols_1.BINDING.Ohno)),
        tslib_1.__param(2, inversify_1.inject(di_symbols_1.BINDING.EventBus)),
        tslib_1.__param(3, inversify_1.inject(di_symbols_1.BINDING.BlockRegistry)),
        tslib_1.__param(4, inversify_1.inject(di_symbols_1.BINDING.OptiCompiler)),
        tslib_1.__param(5, inversify_1.inject(di_symbols_1.BINDING.BlockPool)),
        tslib_1.__metadata("design:paramtypes", [Object, Object, Object, Object, Object, Object])
    ], BlockInterpreterFactoryImpl);
    return BlockInterpreterFactoryImpl;
}());
exports.BlockInterpreterFactoryImpl = BlockInterpreterFactoryImpl;
