"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var _cloneDeep = require("lodash/cloneDeep");
var _isFunction = require("lodash/isFunction");
var inversify_1 = require("inversify");
var block_provider_1 = require("./block_provider");
var di_symbols_1 = require("./di_symbols");
var BlockRegistryImpl = (function () {
    function BlockRegistryImpl(blockly) {
        this.blockly = blockly;
        this.action_specs = {};
        this.responder_infos = {};
        this.domain_functions = {};
        this.blocks_need_restart_when_finished = {};
        this.blocks_need_finish_out_of_run_group = {};
        this.loaded_json_block_configs = {};
        this.loaded_init_function_block_configs = {};
    }
    BlockRegistryImpl.prototype.register_provider = function (p) {
        if (this.blockly != undefined) {
            this.define_blocks(this.blockly, p);
        }
        this.register_runtime_provider(p);
    };
    BlockRegistryImpl.prototype.get_loaded_block_configs = function () {
        return {
            defined_by_json: _cloneDeep(this.loaded_json_block_configs),
            defined_by_init_function: _cloneDeep(this.loaded_init_function_block_configs),
        };
    };
    BlockRegistryImpl.prototype.register_runtime_provider = function (p) {
        var namespace = p.namespace();
        var dom_funs = p.domain_functions();
        for (var fun_name in dom_funs) {
            var fun = dom_funs[fun_name];
            this.domain_functions[block_provider_1.namespaced_id(namespace, fun_name)] = fun;
        }
        var action_types = p.action_types();
        for (var i = 0; i < action_types.length; i++) {
            var action_spec = action_types[i];
            var ns_action_id = block_provider_1.namespaced_id(namespace, action_spec.id);
            this.action_specs[ns_action_id] = action_spec;
            for (var j = 0; j < action_spec.responder_blocks.length; j++) {
                var responder_spec = action_spec.responder_blocks[j];
                var ns_responder_id = block_provider_1.namespaced_id(namespace, responder_spec.id);
                this.responder_infos[ns_responder_id] = {
                    namespace: namespace,
                    action_spec: action_spec,
                    responder_spec: responder_spec,
                };
            }
        }
        if (p.block_metadata === undefined) {
            return;
        }
        if (p.block_metadata.restart_when_finished != undefined) {
            for (var i = 0; i < p.block_metadata.restart_when_finished.length; i++) {
                var block_name = p.block_metadata.restart_when_finished[i];
                var ns_id = block_provider_1.namespaced_id(namespace, block_name);
                this.blocks_need_restart_when_finished[ns_id] = true;
            }
        }
        if (p.block_metadata.finish_out_of_run_group != undefined) {
            for (var i = 0; i < p.block_metadata.finish_out_of_run_group.length; i++) {
                var block_name = p.block_metadata.finish_out_of_run_group[i];
                var ns_id = block_provider_1.namespaced_id(namespace, block_name);
                this.blocks_need_finish_out_of_run_group[ns_id] = true;
            }
        }
    };
    BlockRegistryImpl.prototype.get_action_spec = function (ns_event_id) {
        return this.action_specs[ns_event_id];
    };
    BlockRegistryImpl.prototype.get_spec_of_action = function (event) {
        return this.action_specs[block_provider_1.namespaced_id(event.namespace, event.id)];
    };
    BlockRegistryImpl.prototype.has_responder_type = function (ns_responder_id) {
        return this.responder_infos[ns_responder_id] != undefined;
    };
    BlockRegistryImpl.prototype.get_responder_info = function (ns_responder_id) {
        return this.responder_infos[ns_responder_id];
    };
    BlockRegistryImpl.prototype.get_event_types = function () {
        return Object.keys(this.action_specs);
    };
    BlockRegistryImpl.prototype.get_domain_functions = function () {
        return this.domain_functions;
    };
    BlockRegistryImpl.prototype.block_restart_when_finished = function (ns_id) {
        return this.blocks_need_restart_when_finished[ns_id] != undefined;
    };
    BlockRegistryImpl.prototype.block_finish_out_of_run_group = function (ns_id) {
        return this.blocks_need_finish_out_of_run_group[ns_id] != undefined;
    };
    BlockRegistryImpl.prototype.define_blocks = function (blockly, p) {
        var config = p.config(blockly);
        var _loop_1 = function (name_1) {
            var c = config[name_1];
            var block_name = block_provider_1.namespaced_id(p.namespace(), name_1);
            if (_isFunction(c.init)) {
                var b_cfg_list = this_1.loaded_init_function_block_configs[p.namespace()] || [];
                b_cfg_list.push(name_1);
                this_1.loaded_init_function_block_configs[p.namespace()] = b_cfg_list;
                blockly.Blocks[block_name] = {
                    name: block_name,
                    init: c.init,
                };
            }
            else {
                var b_cfg_dict = this_1.loaded_json_block_configs[p.namespace()] || {};
                b_cfg_dict[name_1] = c;
                this_1.loaded_json_block_configs[p.namespace()] = b_cfg_dict;
                blockly.Blocks[block_name] = {
                    name: block_name,
                    init: function () { this.jsonInit(c); },
                };
            }
        };
        var this_1 = this;
        for (var name_1 in config) {
            _loop_1(name_1);
        }
    };
    BlockRegistryImpl = tslib_1.__decorate([
        inversify_1.injectable(),
        tslib_1.__param(0, inversify_1.inject(di_symbols_1.BINDING.Blockly)), tslib_1.__param(0, inversify_1.optional()),
        tslib_1.__metadata("design:paramtypes", [Object])
    ], BlockRegistryImpl);
    return BlockRegistryImpl;
}());
exports.BlockRegistryImpl = BlockRegistryImpl;
