"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var _defaults = require("lodash/defaults");
var _sortBy = require("lodash/sortBy");
var _identity = require("lodash/identity");
var _flattenDeep = require("lodash/flattenDeep");
var _isString = require("lodash/isString");
var catastrophic_1 = require("catastrophic");
var inversify_1 = require("inversify");
var H = require("../di_interfaces");
var di_symbols_1 = require("../di_symbols");
function peek(stack) {
    if (stack.length == 0) {
        return undefined;
    }
    return stack[stack.length - 1];
}
var SHORT_KEYWORDS = {
    ohno: 'X.AA',
    call_domain_function: 'X.AB',
    report_sync_telling: 'X.AC',
    async_tell: 'X.AD',
    default_target_entity_id: 'X.BA',
    identities: 'X.BB',
    static_data: 'X.BC',
    proc_parameters: 'X.CA',
    proc_return_value: 'X.CB',
    proc_has_return_value: 'X.CC',
    proc_do_return_value: 'X.CD',
    proc_yield_after_call: 'X.CE',
    dynamic_data: 'X.DA',
    program_counter: 'X.DB',
    proc_call: 'X.EA',
    before_expression: 'X.EB',
    after_potential_blocker: 'X.EC',
    increment_program_counter: 'X.ED',
    reset_program_counter: 'X.EE',
    after_iteration: 'X.EF',
    finished: 'X.EG',
};
var PRETTY_KEYWORDS = {
    ohno: 'X.ohno',
    call_domain_function: 'X.call_domain_function',
    report_sync_telling: 'X.report_sync_telling',
    async_tell: 'X.async_tell',
    default_target_entity_id: 'X.default_target_entity_id',
    identities: 'X.identities',
    static_data: 'X.static_data',
    proc_parameters: 'X.proc_parameters',
    proc_return_value: 'X.proc_return_value',
    proc_has_return_value: 'X.proc_has_return_value',
    proc_do_return_value: 'X.proc_do_return_value',
    proc_yield_after_call: 'X.proc_yield_after_call',
    dynamic_data: 'X.dynamic_data',
    program_counter: 'X.program_counter',
    proc_call: 'X.proc_call',
    before_expression: 'X.before_expression',
    after_potential_blocker: 'X.after_potential_blocker',
    increment_program_counter: 'X.increment_program_counter',
    reset_program_counter: 'X.reset_program_counter',
    after_iteration: 'X.after_iteration',
    finished: 'X.finished',
};
var State = (function () {
    function State(ohno, source_map_entity, source_map_rbid, pretty_print_symbols) {
        this.ohno = ohno;
        this.source_map_entity = source_map_entity;
        this.source_map_rbid = source_map_rbid;
        this.pretty_print_symbols = pretty_print_symbols;
        this.current_yield_group = 0;
        this.top_yield_group_id = 0;
        this.top_yield_ids = { 0: 0 };
        this.yield_group_stack = [];
        this.break_yield_reset_stack = [[]];
        this.tell_entity_target_stack = [];
        this.top_break_id = 0;
        this.break_id_stack = [];
        this.n_warp_block_parents = 0;
        this.top_parameter_id = 0;
        this.top_static_data_id = 0;
        this.async_tell_asts = {};
        this.keyword = pretty_print_symbols ? PRETTY_KEYWORDS : SHORT_KEYWORDS;
        this.static_data = pretty_print_symbols ? {} : [];
    }
    State.prototype.gen_parameter_symbol = function (name) {
        this.top_parameter_id++;
        var prefix = this.pretty_print_symbols ? name : '';
        return prefix + this.top_parameter_id;
    };
    State.prototype.set_static = function (name, value) {
        this.top_static_data_id++;
        if (this.pretty_print_symbols) {
            var symbol = [name, this.top_static_data_id].join('');
            this.static_data[symbol] = value;
            return symbol;
        }
        this.static_data[this.top_static_data_id] = value;
        return this.top_static_data_id.toString();
    };
    State.prototype.get_static_data = function () {
        return this.static_data;
    };
    State.prototype.dynamic_data_size = function () {
        return this.top_parameter_id + 1;
    };
    State.prototype.push_yield_level = function (block) {
        this.yield_group_stack.push(this.current_yield_group);
        this.top_yield_group_id++;
        var next_id = this.top_yield_group_id;
        this.current_yield_group = next_id;
        this.top_yield_ids[next_id] = 0;
        var break_yrs = peek(this.break_yield_reset_stack);
        if (break_yrs == undefined) {
            throw this.ohno.compiler.system.popped_empty_yield_reset_stack({
                block: block,
            });
        }
        break_yrs.push(next_id);
        return [
            "switch (", this.keyword.program_counter, "[",
            next_id.toString(),
            "]) {case 0:",
        ];
    };
    State.prototype.pop_yield_level = function (block) {
        var group_id = this.yield_group_stack.pop();
        if (group_id == undefined) {
            throw this.ohno.compiler.system.popped_empty_yield_group_stack({
                block: block,
            });
        }
        this.current_yield_group = group_id;
        var break_yrs = peek(this.break_yield_reset_stack);
        if (break_yrs == undefined) {
            throw this.ohno.compiler.system.popped_empty_yield_group_stack({
                block: block,
            });
        }
        break_yrs.pop();
        return "}";
    };
    State.prototype.push_breakable = function () {
        this.break_yield_reset_stack.push([]);
        var n = this.top_break_id++;
        var break_id = 'B' + n;
        this.break_id_stack.push(break_id);
        return break_id;
    };
    State.prototype.pop_breakable = function () {
        this.break_id_stack.pop();
        this.break_yield_reset_stack.pop();
    };
    State.prototype.get_current_break_reset_statements = function (block) {
        var break_yrs = peek(this.break_yield_reset_stack);
        if (break_yrs == undefined) {
            throw this.ohno.compiler.system.popped_empty_yield_reset_stack({
                block: block,
            });
        }
        var breakers = [];
        for (var i = 0; i < break_yrs.length; i++) {
            var group = break_yrs[i];
            breakers.push(this.keyword.program_counter, "[", group.toString(), "] = 0;");
        }
        return breakers;
    };
    State.prototype.push_target_entity = function (accessor) {
        this.tell_entity_target_stack.push(accessor);
    };
    State.prototype.pop_target_entity = function () {
        this.tell_entity_target_stack.pop();
    };
    State.prototype.current_target_entity = function () {
        var accessor = peek(this.tell_entity_target_stack);
        if (accessor == undefined) {
            return this.keyword.default_target_entity_id;
        }
        return accessor;
    };
    State.prototype.current_break_id = function (block) {
        var id = peek(this.break_id_stack);
        if (id == undefined) {
            throw this.ohno.compiler.user.tried_to_break_outside_of_loop({
                block: block,
            });
        }
        return id;
    };
    State.prototype.create_yield_point = function () {
        this.top_yield_ids[this.current_yield_group]++;
        var n = this.top_yield_ids[this.current_yield_group];
        return [
            this.keyword.increment_program_counter, "('",
            this.source_map_entity,
            "', '",
            this.source_map_rbid,
            "', ",
            this.current_yield_group.toString(),
            ", ",
            n.toString(),
            "); case ", n.toString(), ": ",
        ];
    };
    State.prototype.create_before_expression_yield_point = function (ast) {
        this.top_yield_ids[this.current_yield_group]++;
        var n = this.top_yield_ids[this.current_yield_group];
        return [
            "if (", this.keyword.before_expression, "('",
            ast.id,
            "', '",
            this.source_map_entity,
            "', '",
            this.source_map_rbid,
            "', ",
            this.current_yield_group.toString(),
            ", ",
            n.toString(),
            ")) { return; }",
            "case ", n.toString(), ": ",
        ];
    };
    State.prototype.create_after_potential_blocker_yield_point = function (ast) {
        this.top_yield_ids[this.current_yield_group]++;
        var n = this.top_yield_ids[this.current_yield_group];
        return [
            "if (", this.keyword.after_potential_blocker, "('",
            ast.id,
            "', '",
            this.source_map_entity,
            "', '",
            this.source_map_rbid,
            "', ",
            this.current_yield_group.toString(),
            ", ",
            n.toString(),
            ")) { return; }",
            "case ", n.toString(), ": ",
        ];
    };
    State.prototype.create_iteration_yield_point = function () {
        return [
            "if (", this.keyword.after_iteration, "(", this.is_inside_warp().toString(), ")) { return; }",
        ];
    };
    State.prototype.create_yield_reset_point = function () {
        return [
            this.keyword.reset_program_counter, "('",
            this.source_map_entity,
            "', '",
            this.source_map_rbid,
            "', ",
            this.current_yield_group.toString(),
            ");",
        ];
    };
    State.prototype.create_procedure_yield_point = function () {
        this.top_yield_ids[this.current_yield_group]++;
        var n = this.top_yield_ids[this.current_yield_group];
        return [
            this.keyword.proc_yield_after_call, "(",
            this.current_yield_group.toString(),
            ", ",
            n.toString(),
            "); return;",
            "case ", n.toString(), ": ",
        ];
    };
    State.prototype.get_n_yield_groups = function () {
        return 1 + this.top_yield_group_id;
    };
    State.prototype.add_async_tell_ast = function (ast) {
        this.async_tell_asts[ast.id] = ast;
    };
    State.prototype.get_async_tell_asts = function () {
        return this.async_tell_asts;
    };
    State.prototype.is_inside_warp = function () {
        return this.n_warp_block_parents >= 1;
    };
    State.prototype.enter_warp_block = function () {
        this.n_warp_block_parents++;
    };
    State.prototype.exit_warp_block = function () {
        this.n_warp_block_parents--;
    };
    return State;
}());
var OptiCompilerImpl = (function () {
    function OptiCompilerImpl(block_pool, block_registry, event_bus, log, ohno, prng_factory, u, block, config) {
        var _this = this;
        this.block_pool = block_pool;
        this.block_registry = block_registry;
        this.event_bus = event_bus;
        this.log = log;
        this.ohno = ohno;
        this.prng_factory = prng_factory;
        this.u = u;
        this.block = block;
        this.config = config;
        this.deterministic = u.config.get().deterministic != undefined;
        var configure = function () {
            _this.should_pretty_print = false;
            var cfg = _this.config.get().opti_compiler;
            if (cfg != undefined) {
                _this.should_pretty_print = cfg.pretty_print;
            }
            _this.keyword = _this.should_pretty_print ? PRETTY_KEYWORDS : SHORT_KEYWORDS;
        };
        configure();
        this.event_bus.system.config_updated.immediate.sub(configure);
    }
    OptiCompilerImpl.prototype.create_catch_all_error = function (location, error, ast) {
        var metadata = {
            caught_at: location,
        };
        if (ast != undefined) {
            metadata.block = ast;
        }
        if (error instanceof catastrophic_1.Catastrophe) {
            error.annotation = error.annotation || {};
            _defaults(error.annotation, metadata);
            return error;
        }
        return this.ohno.compiler.system.unknown_compiler_error(error, metadata);
    };
    OptiCompilerImpl.prototype.compile = function (source_map_entity, source_map_rbid, interpreter_id, ast) {
        var code;
        var static_data;
        var dynamic_data_size;
        var n_yield_groups;
        var async_tell_asts;
        try {
            var state = new State(this.ohno, source_map_entity, source_map_rbid, this.should_pretty_print);
            var s = this.compile_statement(state, ast);
            var r = this.wrap_with_top_level(s);
            state = r.state;
            code = _flattenDeep(r.statements).join('');
            static_data = state.get_static_data();
            dynamic_data_size = state.dynamic_data_size();
            n_yield_groups = state.get_n_yield_groups();
            async_tell_asts = state.get_async_tell_asts();
        }
        catch (e) {
            var metadata = {
                block: ast,
                interpreter_id: interpreter_id,
                source_map_rbid: source_map_rbid,
                source_entity_id: source_map_entity,
                proc_id: (this.block.is.procedures_defnoreturn(ast)) ? ast.procedure_name : undefined,
            };
            if (e instanceof catastrophic_1.Catastrophe) {
                var annotation = e.annotation || {};
                _defaults(annotation, metadata);
                if (annotation.block_id == undefined && annotation.block != undefined) {
                    annotation.block_id = annotation.block.id;
                }
                if (annotation.block_type == undefined && annotation.block != undefined) {
                    annotation.block_type = annotation.block.type;
                }
                e.annotation = annotation;
                throw e;
            }
            throw this.ohno.compiler.system.unknown_compiler_error(e, metadata);
        }
        var optifun;
        try {
            optifun = new Function('X', code)();
        }
        catch (e) {
            throw this.ohno.compiler.system.constructed_bad_javascript(e, {
                root_block_id: source_map_rbid,
                entity_id: source_map_entity,
                invalid_script: code,
            });
        }
        return {
            script: optifun,
            static_data: static_data,
            dynamic_data_size: dynamic_data_size,
            n_yield_groups: n_yield_groups,
            async_tell_asts: async_tell_asts,
        };
    };
    OptiCompilerImpl.prototype.wrap_with_top_level = function (program) {
        var program_statements = program.statements;
        var state = program.state;
        var last_yield = state.create_yield_point();
        var statements = [
            "return function optifun(X) {",
            "switch (", this.keyword.program_counter, "[0]) {",
            "case 0: ",
            program.statements,
            last_yield,
            this.keyword.finished, "();",
            "}}",
        ];
        return {
            statements: statements,
            state: state,
        };
    };
    OptiCompilerImpl.prototype.compile_statement = function (state, ast) {
        try {
            if (ast == undefined) {
                return {
                    state: state,
                    statements: '',
                };
            }
            var res = [];
            res.push(state.create_before_expression_yield_point(ast));
            var omit_semicolon = false;
            var omit_next = false;
            if (ast.disabled) {
                if (ast.next_block != undefined && omit_next == false) {
                    var r = this.compile_statement(state, ast.next_block);
                    state = r.state;
                    res.push(r.statements);
                }
                return {
                    state: state,
                    statements: res,
                };
            }
            if (this.block.is.repeat_forever(ast)) {
                omit_semicolon = true;
                var r = this.compile_repeat_forever(state, ast);
                state = r.state;
                res.push(r.statements);
            }
            else if (this.block.is.repeat_forever_until(ast)) {
                omit_semicolon = true;
                var r = this.compile_repeat_forever_until(state, ast);
                state = r.state;
                res.push(r.statements);
            }
            else if (this.block.is.wait_until(ast)) {
                omit_semicolon = true;
                var r = this.compile_wait_until(state, ast);
                state = r.state;
                res.push(r.statements);
            }
            else if (this.block.is.repeat_n_times(ast)) {
                omit_semicolon = true;
                var r = this.compile_repeat_n_times(state, ast);
                state = r.state;
                res.push(r.statements);
            }
            else if (this.block.is.break(ast)) {
                var r = this.compile_break(state, ast);
                state = r.state;
                res.push(r.statements);
            }
            else if (this.block.is.procedures_defnoreturn(ast)) {
                var body = ast.child_block[0];
                if (body == undefined) {
                    omit_semicolon = true;
                }
                else {
                    var r = this.compile_statement(state, body);
                    state = r.state;
                    res.push(r.statements);
                }
            }
            else if (this.block.is.procedures_callnoreturn(ast)) {
                var r = this.compile_procedure_call(state, ast);
                state = r.state;
                res.push(r.pre);
            }
            else if (this.block.is.cond_block(ast)) {
                omit_semicolon = true;
                var r = this.compile_conditional(state, ast);
                state = r.state;
                res.push(r.statements);
            }
            else if (this.block.is.async_tell(ast)) {
                omit_semicolon = true;
                var r = this.compile_async_tell(state, ast);
                state = r.state;
                res.push(r.statements);
            }
            else if (this.block.is.sync_tell(ast)) {
                omit_semicolon = true;
                var r = this.compile_sync_tell(state, ast);
                state = r.state;
                res.push(r.statements);
            }
            else if (this.block.is.warp(ast)) {
                omit_semicolon = true;
                var r = this.compile_warp(state, ast);
                state = r.state;
                res.push(r.statements);
            }
            else if (this.block.is.event_block(ast)) {
                omit_semicolon = true;
                var r = this.compile_event(state, ast);
                state = r.state;
                res.push(r.statements);
            }
            else if (this.block.is.responder_block(ast)) {
                omit_semicolon = true;
                var r = this.compile_responder(state, ast);
                state = r.state;
                res.push(r.statements);
            }
            else if (this.block.is.procedures_return_value(ast)) {
                omit_semicolon = true;
                omit_next = true;
                var r = this.compile_procedure_return_value(state, ast);
                state = r.state;
                res.push(r.statements);
            }
            else {
                var r = this.compile_expression(state, ast);
                state = r.state;
                res.push(r.pre);
                res.push(r.expr);
            }
            if (res.length != 0 && omit_semicolon == false) {
                res.push(";");
            }
            if (ast.next_block != undefined && omit_next == false) {
                var r = this.compile_statement(state, ast.next_block);
                state = r.state;
                res.push(r.statements);
            }
            return {
                state: state,
                statements: res,
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_statement', e, ast);
        }
    };
    OptiCompilerImpl.prototype.compile_expression = function (state, ast) {
        if (ast.disabled) {
            throw this.ohno.compiler.user.disabled_param(new Error(), {
                block: ast,
            });
        }
        try {
            if (this.block.is.atomic_type(ast.type)) {
                var static_value = ast.params[Object.keys(ast.params)[0]];
                var static_symbol = state.set_static(ast.type + "__" + ast.id, static_value);
                return {
                    pre: '',
                    expr: this.opt_static(static_symbol),
                    state: state,
                };
            }
        }
        catch (e) {
            throw this.ohno.compiler.user.error_constructing_value_from_atomic_block(e, {
                block: ast,
            });
        }
        try {
            if (this.block.is.logic_empty(ast)) {
                return this.compile_logic_empty(state, ast);
            }
            var before_expression_yield = state.create_before_expression_yield_point(ast);
            var res = void 0;
            if (this.block.is.procedures_callreturn(ast)) {
                res = this.compile_procedure_call(state, ast);
            }
            else if (this.block.is.procedures_parameter(ast)) {
                res = this.compile_procedure_parameter(state, ast);
            }
            else if (this.block.is.domain_block(ast)) {
                res = this.compile_domain_functions(state, ast);
            }
            else {
                var block = ast;
                throw this.ohno.compiler.system.unknown_expression({
                    unknown_expression_type: block.type,
                    block: block,
                });
            }
            return {
                state: res.state,
                pre: [before_expression_yield, res.pre],
                expr: res.expr,
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_expression', e, ast);
        }
    };
    OptiCompilerImpl.prototype.compile_conditional = function (state, ast) {
        var _this = this;
        try {
            if (ast.child_block.length == 0 || ast.conditions.length == 0) {
                return {
                    state: state,
                    statements: '',
                };
            }
            var r = [];
            var matched_id = state.gen_parameter_symbol("if_matched");
            var matched_param = this.opt_params(matched_id);
            r.push(state.create_yield_point());
            r.push(matched_param, " = false;");
            var compile_branch = function (branch) {
                if (branch == undefined) {
                    return [];
                }
                var r_branch = [];
                r_branch.push(state.push_yield_level(ast));
                var branch_comp = _this.compile_statement(state, branch);
                state = branch_comp.state;
                r_branch.push(branch_comp.statements);
                r_branch.push(state.create_yield_reset_point());
                r_branch.push(state.pop_yield_level(ast));
                return r_branch;
            };
            for (var i = 0; i < ast.conditions.length; i++) {
                var cond = ast.conditions[i];
                var branch = ast.child_block[i];
                if (cond == undefined) {
                    continue;
                }
                var cond_id = state.gen_parameter_symbol("if_cond" + i + "_");
                var cond_param = this.opt_params(cond_id);
                var cond_push = state.push_yield_level(ast);
                var r_cond = this.compile_expression(state, cond);
                state = r_cond.state;
                var cond_reset = state.create_yield_reset_point();
                var cond_pop = state.pop_yield_level(ast);
                var pre_yield = state.create_yield_point();
                var r_branch = compile_branch(branch);
                var post_yield = state.create_yield_point();
                r.push("if (!", matched_param, ") {", cond_push, r_cond.pre, cond_reset, cond_pop, "}");
                r.push(cond_param, " = !", matched_param, " && ", r_cond.expr, ";");
                r.push(matched_param, " = ", matched_param, " || ", cond_param, ";");
                r.push(pre_yield);
                r.push("if (", cond_param, ") {", r_branch, "}");
                r.push(post_yield);
            }
            var else_branch = ast.child_block[ast.conditions.length];
            if (else_branch != undefined) {
                var r_branch = compile_branch(else_branch);
                r.push("if (!", matched_param, ") {", r_branch, "}");
            }
            return {
                state: state,
                statements: r,
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_conditional', e, ast);
        }
    };
    OptiCompilerImpl.prototype.compile_sync_tell = function (state, ast) {
        try {
            if (ast.child_block[0] == undefined) {
                return {
                    state: state,
                    statements: '',
                };
            }
            var params = this.compile_params(state, "__sync_tell_" + ast.id + "__", ast.params);
            state = params.state;
            var target_id_param = [this.opt_params(params.dynamic_data_id), "['sprite']"];
            state.push_target_entity(target_id_param);
            var body = this.compile_statement(state, ast.child_block[0]);
            state = body.state;
            state.pop_target_entity();
            var statements = [
                params.statements,
                this.keyword.report_sync_telling, "('", ast.id, "', ", target_id_param, ");",
                body.statements,
            ];
            return {
                state: state,
                statements: statements,
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_sync_tell', e, ast);
        }
    };
    OptiCompilerImpl.prototype.compile_warp = function (state, ast) {
        try {
            if (ast.child_block[0] == undefined) {
                return {
                    state: state,
                    statements: '',
                };
            }
            state.enter_warp_block();
            var body = this.compile_statement(state, ast.child_block[0]);
            state = body.state;
            state.exit_warp_block();
            return {
                state: state,
                statements: body.statements,
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_warp', e, ast);
        }
    };
    OptiCompilerImpl.prototype.compile_async_tell = function (state, ast) {
        try {
            var tell_child = ast.child_block[0];
            if (tell_child == undefined) {
                return {
                    state: state,
                    statements: '',
                };
            }
            state.add_async_tell_ast(tell_child);
            var params = this.compile_params(state, "__async_tell_" + ast.id + "__", ast.params);
            state = params.state;
            var target_id_param = [this.opt_params(params.dynamic_data_id), "['sprite']"];
            var statements = [
                params.statements,
                this.keyword.async_tell,
                "('",
                tell_child.id,
                "', ",
                state.current_target_entity(),
                ", ",
                target_id_param,
                ", ",
                state.is_inside_warp().toString(),
                ");",
            ];
            return {
                state: state,
                statements: statements,
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_async_tell', e, ast);
        }
    };
    OptiCompilerImpl.prototype.compile_repeat_forever = function (state, ast) {
        try {
            if (ast.child_block[0] == undefined) {
                return {
                    state: state,
                    statements: '',
                };
            }
            var break_id = state.push_breakable();
            var push = state.push_yield_level(ast);
            var r = this.compile_statement(state, ast.child_block[0]);
            state = r.state;
            var yield_reset = state.create_yield_reset_point();
            var iter_yield = state.create_iteration_yield_point();
            var pop = state.pop_yield_level(ast);
            state.pop_breakable();
            var statements = [
                break_id, ": while (true) {",
                push,
                r.statements,
                yield_reset,
                iter_yield,
                pop,
                "}",
            ];
            return {
                state: state,
                statements: statements,
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_repeat_forever', e, ast);
        }
    };
    OptiCompilerImpl.prototype.compile_repeat_forever_until = function (state, ast) {
        try {
            if (ast.child_block[0] == undefined) {
                return {
                    state: state,
                    statements: '',
                };
            }
            var break_id = state.push_breakable();
            var push = state.push_yield_level(ast);
            var params = this.compile_params(state, "__repeat_forever_until_" + ast.id + "__", ast.params);
            state = params.state;
            var cond_param = [this.opt_params(params.dynamic_data_id), "['condition']"];
            var break_reset = state.create_yield_reset_point();
            var r = this.compile_statement(state, ast.child_block[0]);
            state = r.state;
            var yield_reset = state.create_yield_reset_point();
            var iter_yield = state.create_iteration_yield_point();
            var pop = state.pop_yield_level(ast);
            state.pop_breakable();
            var statements = [
                break_id, ": while (true) {",
                push,
                params.statements,
                "if (", cond_param, ") {", break_reset, "break ", break_id, ";}",
                r.statements,
                yield_reset,
                iter_yield,
                pop,
                "}",
            ];
            return {
                state: state,
                statements: statements,
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_repeat_forever_until', e, ast);
        }
    };
    OptiCompilerImpl.prototype.compile_repeat_n_times = function (state, ast) {
        try {
            if (ast.child_block[0] == undefined) {
                return {
                    state: state,
                    statements: '',
                };
            }
            var params = this.compile_params(state, "__repeat_n_times_" + ast.id + "__", ast.params);
            state = params.state;
            var break_id = state.push_breakable();
            var push = state.push_yield_level(ast);
            var iter_param = [this.opt_params(params.dynamic_data_id), "['times']"];
            var break_reset = state.create_yield_reset_point();
            var r = this.compile_statement(state, ast.child_block[0]);
            state = r.state;
            var yield_reset = state.create_yield_reset_point();
            var iter_yield = state.create_iteration_yield_point();
            var pop = state.pop_yield_level(ast);
            state.pop_breakable();
            var statements = [
                params.statements,
                break_id, ": while (true) {",
                push,
                "if (", iter_param, " <= 0) {", break_reset, "break ", break_id, ";}",
                iter_param, "--;",
                r.statements,
                yield_reset,
                iter_yield,
                pop,
                "}",
            ];
            return {
                state: state,
                statements: statements,
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_repeat_n_times', e, ast);
        }
    };
    OptiCompilerImpl.prototype.compile_wait_until = function (state, ast) {
        try {
            var break_id = state.push_breakable();
            var push = state.push_yield_level(ast);
            var params = this.compile_params(state, "__wait_until_" + ast.id + "__", ast.params);
            state = params.state;
            var cond_param = [this.opt_params(params.dynamic_data_id), "['condition']"];
            var break_reset = state.create_yield_reset_point();
            var yield_reset = state.create_yield_reset_point();
            var iter_yield = state.create_iteration_yield_point();
            var pop = state.pop_yield_level(ast);
            state.pop_breakable();
            var statements = [
                break_id, ": while (true) {",
                push,
                params.statements,
                "if (", cond_param, ") {", break_reset, "break ", break_id, ";}",
                yield_reset,
                iter_yield,
                pop,
                "}",
            ];
            return {
                state: state,
                statements: statements,
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_wait_until', e, ast);
        }
    };
    OptiCompilerImpl.prototype.compile_break = function (state, ast) {
        try {
            var statements = [
                state.get_current_break_reset_statements(ast),
                state.create_yield_reset_point(),
                "break ", state.current_break_id(ast), ";",
            ];
            return {
                state: state,
                statements: statements,
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_break', e, ast);
        }
    };
    OptiCompilerImpl.prototype.compile_logic_empty = function (state, ast) {
        try {
            return {
                state: state,
                pre: [],
                expr: ['false'],
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_logic_empty', e, ast);
        }
    };
    OptiCompilerImpl.prototype.compile_domain_functions = function (state, ast) {
        try {
            var params = this.compile_params(state, "__domain_block_" + ast.type + "_" + ast.id + "__", ast.params);
            var res_params = [];
            res_params.push(params.statements);
            var wrap_id = state.gen_parameter_symbol("domain_wrap");
            var wrap_param = this.opt_params(wrap_id);
            var res_expr = [
                this.keyword.call_domain_function, "('",
                wrap_id,
                "', '",
                ast.type,
                "', '",
                ast.id,
                "', ",
                this.opt_params(params.dynamic_data_id),
                ", ",
                state.current_target_entity(),
                ");",
            ];
            var res_expr_wrap = [
                res_expr,
                state.create_after_potential_blocker_yield_point(ast),
            ];
            return {
                state: params.state,
                pre: [res_params, res_expr_wrap],
                expr: wrap_param,
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_domain_functions', e, ast);
        }
    };
    OptiCompilerImpl.prototype.compile_event = function (state, ast) {
        try {
            var triggered_id = state.gen_parameter_symbol("if_triggered");
            var triggered_param = this.opt_params(triggered_id);
            var c = this.compile_domain_functions(state, ast);
            state = c.state;
            var yield_after_test = state.create_yield_point();
            var yield_before_enter = state.create_yield_point();
            var push = state.push_yield_level(ast);
            var b = this.compile_statement(state, ast.child_block[0]);
            state = b.state;
            var yield_reset = state.create_yield_reset_point();
            var pop = b.state.pop_yield_level(ast);
            return {
                state: state,
                statements: [
                    c.pre,
                    yield_after_test,
                    triggered_param, " = ", c.expr, ";",
                    yield_before_enter,
                    "if (", triggered_param, ") {",
                    push,
                    b.statements,
                    yield_reset,
                    pop,
                    "}",
                ],
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_event', e, ast);
        }
    };
    OptiCompilerImpl.prototype.compile_responder = function (state, ast) {
        try {
            var b = this.compile_statement(state, ast.child_block[0]);
            state = b.state;
            return {
                state: state,
                statements: [
                    b.statements,
                ],
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_responder', e, ast);
        }
    };
    OptiCompilerImpl.prototype.compile_procedure_call = function (state, ast) {
        try {
            var params = this.compile_params(state, "__" + ast.type + "_" + ast.id + "__", ast.params);
            state = params.state;
            var res_pre = [];
            res_pre.push(params.statements);
            res_pre.push(state.create_yield_point());
            res_pre.push([
                this.keyword.proc_call, "('",
                ast.procedure_name,
                "', ",
                state.current_target_entity(),
                ", ",
                this.opt_params(params.dynamic_data_id),
                ", '",
                ast.id,
                "', ",
                state.is_inside_warp().toString(),
                ");",
            ]);
            res_pre.push(state.create_procedure_yield_point());
            var res_expr = [
                this.keyword.proc_return_value,
            ];
            return {
                state: state,
                pre: res_pre,
                expr: res_expr,
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_procedure_call', e, ast);
        }
    };
    OptiCompilerImpl.prototype.compile_procedure_parameter = function (state, ast) {
        var root_block = this.u.block.get_first_ancestor_satisfying(ast, function (cb) { return cb.parent_block == undefined; });
        if (root_block == undefined) {
            throw this.ohno.compiler.system.could_not_find_root_block({
                block: ast,
            });
        }
        if (!this.u.block.is.procedures_defnoreturn(root_block)) {
            throw this.ohno.compiler.user.procedure_parameter_outside({
                block: ast,
            });
        }
        var param_name;
        if (ast.params == undefined
            || ast.params.param_name == undefined
            || !_isString(ast.params.param_name)) {
            throw this.ohno.compiler.system.could_not_find_procedure_parameter_name({
                block: ast,
            });
        }
        else {
            param_name = ast.params.param_name;
        }
        if (!root_block.params[param_name]) {
            throw this.ohno.compiler.user.procedure_no_such_parameter({
                param_name: param_name,
                block: ast,
            });
        }
        try {
            var params = this.compile_params(state, "__" + ast.type + "_" + ast.id + "__", ast.params);
            state = params.state;
            var proc_param_name = [this.opt_params(params.dynamic_data_id), "['param_name']"];
            var res_expr = [
                this.keyword.proc_parameters, "[", proc_param_name, "]",
            ];
            return {
                state: state,
                pre: [
                    params.statements,
                    "if (", res_expr, " == undefined) {throw ",
                    this.keyword.ohno, ".user.proc_parameter_without_value({",
                    "\"block_id\": '", ast.id, "'});}",
                ],
                expr: res_expr,
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_procedure_parameter', e, ast);
        }
    };
    OptiCompilerImpl.prototype.compile_procedure_return_value = function (state, ast) {
        var root_block = this.u.block.get_first_ancestor_satisfying(ast, function (cb) { return cb.parent_block == undefined; });
        if (ast.params['VALUE'] === undefined) {
            throw this.ohno.compiler.user.procedure_return_empty({
                block: ast,
            });
        }
        if (root_block === undefined || !this.u.block.is.procedures_defnoreturn(root_block)) {
            throw this.ohno.compiler.user.procedure_return_outside({
                block: ast,
            });
        }
        try {
            var params = this.compile_params(state, "__" + ast.type + "_" + ast.id + "__", ast.params);
            state = params.state;
            var return_value = [this.opt_params(params.dynamic_data_id), "['VALUE']"];
            var res = [
                params.statements,
                this.keyword.proc_do_return_value, "('", ast.id, "', ", return_value, "); return;",
            ];
            return {
                state: state,
                statements: res,
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_procedure_return_value', e, ast);
        }
    };
    OptiCompilerImpl.prototype.compile_params = function (state, name, params) {
        try {
            var statements = [state.create_yield_point()];
            var param_collection_id = state.gen_parameter_symbol(name + "_params");
            statements.push(this.opt_create_params(param_collection_id));
            var param_keys = Object.keys(params);
            if (param_keys.length == 0) {
                return {
                    state: state,
                    dynamic_data_id: param_collection_id,
                    statements: statements,
                };
            }
            if (this.deterministic) {
                param_keys = _sortBy(param_keys, _identity);
            }
            for (var i = 0; i < param_keys.length; i++) {
                var p_id = param_keys[i];
                var param = params[p_id];
                if (this.block.is.compiled_block(param)) {
                    var expr = this.compile_expression(state, param);
                    state = expr.state;
                    statements.push(expr.pre);
                    statements.push(state.create_yield_point());
                    statements.push(this.opt_set_param(param_collection_id, p_id, expr.expr));
                }
                else {
                    statements.push(state.create_yield_point());
                    var static_symbol = state.set_static(name + '_' + p_id, param);
                    statements.push(this.opt_set_param(param_collection_id, p_id, this.opt_static(static_symbol)));
                }
            }
            statements.push(state.create_yield_point());
            return {
                state: state,
                dynamic_data_id: param_collection_id,
                statements: statements,
            };
        }
        catch (e) {
            throw this.create_catch_all_error('compile_params', e);
        }
    };
    OptiCompilerImpl.prototype.opt_static = function (key) {
        if (this.should_pretty_print) {
            return [this.keyword.static_data, "['", key, "']"];
        }
        return [this.keyword.static_data, "[", key, "]"];
    };
    OptiCompilerImpl.prototype.opt_params = function (key) {
        if (this.should_pretty_print) {
            return [this.keyword.dynamic_data, "['", key, "']"];
        }
        return [this.keyword.dynamic_data, "[", key, "]"];
    };
    OptiCompilerImpl.prototype.opt_create_params = function (key) {
        return [this.opt_params(key), " = {};"];
    };
    OptiCompilerImpl.prototype.opt_set_param = function (key, param, value) {
        return [this.opt_params(key), "['", param, "'] = ", value, ";"];
    };
    OptiCompilerImpl = tslib_1.__decorate([
        inversify_1.injectable(),
        tslib_1.__param(0, inversify_1.inject(di_symbols_1.BINDING.BlockPool)),
        tslib_1.__param(1, inversify_1.inject(di_symbols_1.BINDING.BlockRegistry)),
        tslib_1.__param(2, inversify_1.inject(di_symbols_1.BINDING.EventBus)),
        tslib_1.__param(3, inversify_1.inject(di_symbols_1.BINDING.Log)),
        tslib_1.__param(4, inversify_1.inject(di_symbols_1.BINDING.Ohno)),
        tslib_1.__param(5, inversify_1.inject(di_symbols_1.BINDING.PRNGFactory)),
        tslib_1.__param(6, inversify_1.inject(di_symbols_1.BINDING.Util)),
        tslib_1.__param(7, inversify_1.inject(di_symbols_1.BINDING.BlockUtil)),
        tslib_1.__param(8, inversify_1.inject(di_symbols_1.BINDING.Config)),
        tslib_1.__metadata("design:paramtypes", [Object, Object, Object, Object, Object, Object, Object, Object, Object])
    ], OptiCompilerImpl);
    return OptiCompilerImpl;
}());
exports.OptiCompilerImpl = OptiCompilerImpl;
