"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var _remove = require("lodash/remove");
var inversify_1 = require("inversify");
var H = require("./di_interfaces");
var di_symbols_1 = require("./di_symbols");
var RunStatus;
(function (RunStatus) {
    RunStatus[RunStatus["Running"] = 0] = "Running";
    RunStatus[RunStatus["Stopped"] = 1] = "Stopped";
})(RunStatus || (RunStatus = {}));
var RuntimeDataImpl = (function () {
    function RuntimeDataImpl(state_store, event_bus, u, config) {
        var _this = this;
        this.state_store = state_store;
        this.event_bus = event_bus;
        this.u = u;
        this.config = config;
        this._clone_id_2_original_id = {};
        this.original_id_2_clone_id_list = {};
        this._is_mirror = {};
        this.run_status = RunStatus.Stopped;
        this.arbitrary_data = {};
        this.interpreter_data = {};
        this.running_blocks_map = {};
        this.per_entity_clone_limit = this.config.get().per_entity_clone_limit;
        this.should_report_current_running_block = this.config.get().should_report_current_running_block;
        this.event_bus.system.config_updated.immediate.sub(function () {
            _this.per_entity_clone_limit = _this.config.get().per_entity_clone_limit;
            _this.should_report_current_running_block = _this.config.get().should_report_current_running_block;
        });
    }
    RuntimeDataImpl.prototype.clear = function () {
        this.run_status = RunStatus.Stopped;
        this._clone_id_2_original_id = {};
        this.arbitrary_data = {};
        this.interpreter_data = {};
        this._is_mirror = {};
        this.original_id_2_clone_id_list = {};
        if (!this.should_report_current_running_block) {
            return;
        }
        this.clear_running_blocks_map();
    };
    RuntimeDataImpl.prototype.set_mirror = function (interpreter_id) {
        this._is_mirror[interpreter_id] = true;
    };
    RuntimeDataImpl.prototype.is_mirror = function (interpreter_id) {
        return this._is_mirror[interpreter_id] == true;
    };
    RuntimeDataImpl.prototype.set_arbitrary_data = function (key, value) {
        this.arbitrary_data[key] = value;
    };
    RuntimeDataImpl.prototype.get_arbitrary_data = function (key) {
        return this.arbitrary_data[key];
    };
    RuntimeDataImpl.prototype.set_interpreter_data = function (interpreter_id, key, value) {
        if (this.interpreter_data[interpreter_id] == undefined) {
            this.interpreter_data[interpreter_id] = {};
        }
        this.interpreter_data[interpreter_id][key] = value;
    };
    RuntimeDataImpl.prototype.get_interpreter_data = function (interpreter_id, key) {
        var value = this.interpreter_data[interpreter_id][key];
        return value;
    };
    RuntimeDataImpl.prototype.dispose_interpreter_data = function (interpreter_id) {
        this.interpreter_data[interpreter_id] = {};
        delete (this._is_mirror[interpreter_id]);
        if (this.should_report_current_running_block && (this.running_blocks_map[interpreter_id] !== undefined)) {
            this.set_finished_block(this.running_blocks_map[interpreter_id]);
            delete (this.running_blocks_map[interpreter_id]);
        }
    };
    RuntimeDataImpl.prototype.report_variable_updated = function (var_id, new_value) {
        this.event_bus.runtime_data.variable_update.send({
            var_id: var_id,
            new_value: new_value,
        });
    };
    RuntimeDataImpl.prototype.report_list_updated = function (list_id, new_value) {
        this.event_bus.runtime_data.list_update.send({
            list_id: list_id,
            new_value: new_value,
        });
    };
    RuntimeDataImpl.prototype.report_entity_variable_updated = function (var_id, new_value, entity_id) {
        this.event_bus.runtime_data.entity_variable_update.send({
            var_id: var_id,
            new_value: new_value,
            entity_id: entity_id,
        });
    };
    RuntimeDataImpl.prototype.report_entity_list_updated = function (list_id, new_value, entity_id) {
        this.event_bus.runtime_data.entity_list_update.send({
            list_id: list_id,
            new_value: new_value,
            entity_id: entity_id,
        });
    };
    RuntimeDataImpl.prototype.is_running = function () {
        return this.run_status === RunStatus.Running;
    };
    RuntimeDataImpl.prototype.is_stopped = function () {
        return this.run_status === RunStatus.Stopped;
    };
    RuntimeDataImpl.prototype.set_running = function () {
        this.run_status = RunStatus.Running;
    };
    RuntimeDataImpl.prototype.set_stopped = function () {
        this.run_status = RunStatus.Stopped;
    };
    RuntimeDataImpl.prototype.clone_id_2_original_id = function (entity_id) {
        return this._clone_id_2_original_id[entity_id];
    };
    RuntimeDataImpl.prototype.get_sprite_clones = function (sprite_id) {
        return this.original_id_2_clone_id_list[sprite_id];
    };
    RuntimeDataImpl.prototype.clone_created = function (original_entity_id, clone_entity_id) {
        this._clone_id_2_original_id[clone_entity_id] = original_entity_id;
        if (!this.original_id_2_clone_id_list[original_entity_id]) {
            this.original_id_2_clone_id_list[original_entity_id] = [];
        }
        this.original_id_2_clone_id_list[original_entity_id].push(clone_entity_id);
        if (this.per_entity_clone_limit == undefined) {
            return;
        }
        var clone_list = this.original_id_2_clone_id_list[original_entity_id];
        var n_too_many = clone_list.length - this.per_entity_clone_limit;
        for (var i = 0; i < n_too_many; i++) {
            this.event_bus.clones.dispose_clone.send(clone_list[i]);
        }
    };
    RuntimeDataImpl.prototype.remove_clone_lookups = function (removed_id) {
        var original_id = this._clone_id_2_original_id[removed_id];
        if (original_id == undefined) {
            return;
        }
        delete (this._clone_id_2_original_id[removed_id]);
        var clone_list = this.original_id_2_clone_id_list[original_id];
        _remove(clone_list, function (clone_id) { return clone_id === removed_id; });
        if (clone_list.length === 0) {
            delete (this.original_id_2_clone_id_list[original_id]);
        }
    };
    RuntimeDataImpl.prototype.entity_disposed = function (removed_id) {
        this.remove_clone_lookups(removed_id);
        this.event_bus.runtime_data.entity_dispose.send({ entity_id: removed_id });
    };
    RuntimeDataImpl.prototype.get_action_state_value = function (params) {
        return this.state_store.get_action_state_value(params);
    };
    RuntimeDataImpl.prototype.set_running_block = function (interpreter_id, running_block_id) {
        if (this.running_blocks_map[interpreter_id] === running_block_id) {
            return;
        }
        if (this.running_blocks_map[interpreter_id] !== undefined) {
            this.event_bus.runtime_data.block_finished.send(this.running_blocks_map[interpreter_id]);
        }
        this.running_blocks_map[interpreter_id] = running_block_id;
        this.event_bus.runtime_data.block_running.send(running_block_id);
    };
    RuntimeDataImpl.prototype.set_finished_block = function (finished_block_id) {
        this.event_bus.runtime_data.block_finished.send(finished_block_id);
    };
    RuntimeDataImpl.prototype.clear_running_blocks_map = function () {
        var running_interpreters = Object.keys(this.running_blocks_map);
        for (var i = 0; i < running_interpreters.length; i++) {
            var interpreter_id = running_interpreters[i];
            this.set_finished_block(this.running_blocks_map[interpreter_id]);
        }
        this.running_blocks_map = {};
    };
    RuntimeDataImpl = tslib_1.__decorate([
        inversify_1.injectable(),
        tslib_1.__param(0, inversify_1.inject(di_symbols_1.BINDING.ActionStateStore)),
        tslib_1.__param(1, inversify_1.inject(di_symbols_1.BINDING.EventBus)),
        tslib_1.__param(2, inversify_1.inject(di_symbols_1.BINDING.Util)),
        tslib_1.__param(3, inversify_1.inject(di_symbols_1.BINDING.Config)),
        tslib_1.__metadata("design:paramtypes", [Object, Object, Object, Object])
    ], RuntimeDataImpl);
    return RuntimeDataImpl;
}());
exports.RuntimeDataImpl = RuntimeDataImpl;
