"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var _filter = require("lodash/filter");
var _includes = require("lodash/includes");
var _remove = require("lodash/remove");
var inversify_1 = require("inversify");
var di_symbols_1 = require("./di_symbols");
var H = require("./di_interfaces");
var TaskManagerImpl = (function () {
    function TaskManagerImpl(event_bus, runtime_data, u) {
        this.event_bus = event_bus;
        this.runtime_data = runtime_data;
        this.u = u;
        this.tasks = [];
        this.tasks_needing_dispose = [];
        this.interpreter_n_blockers_lookup = {};
        var d = this.u.config.get().deterministic;
        if (d == undefined) {
            this.update_function = this.wall_clock_task_update.bind(this);
        }
        else {
            this.ms_per_update = d.seconds_per_update * 1000;
            this.update_function = this.deterministic_task_update.bind(this);
        }
    }
    TaskManagerImpl.prototype.clear = function () {
        this.tasks = [];
        this.tasks_needing_dispose = [];
        this.interpreter_n_blockers_lookup = {};
    };
    TaskManagerImpl.prototype.update_dispose = function () {
        var _this = this;
        var removed_tasks = _remove(this.tasks, function (t) { return _includes(_this.tasks_needing_dispose, t.id); });
        for (var i = 0; i < removed_tasks.length; i++) {
            var t = removed_tasks[i];
            var interp_id = t.interpreter_id;
            if (t.blocking && this.interpreter_n_blockers_lookup[interp_id] != undefined) {
                this.interpreter_n_blockers_lookup[interp_id]--;
                if (this.interpreter_n_blockers_lookup[interp_id] < 1) {
                    delete (this.interpreter_n_blockers_lookup[interp_id]);
                }
            }
        }
        this.tasks_needing_dispose = [];
    };
    TaskManagerImpl.prototype.update = function () {
        if (this.runtime_data.is_stopped()) {
            this.clear();
            return;
        }
        for (var i = 0; i < this.tasks.length; i++) {
            this.update_function(this.tasks[i]);
        }
    };
    TaskManagerImpl.prototype.add_task = function (id, t) {
        var _this = this;
        var task = t;
        task.id = id;
        this.tasks.push(task);
        if (task.blocking) {
            var interp_id = t.interpreter_id;
            this.interpreter_n_blockers_lookup[interp_id] = this.interpreter_n_blockers_lookup[interp_id] || 0;
            this.interpreter_n_blockers_lookup[interp_id]++;
        }
        return {
            stop: function () { return _this.dispose_task(task.id); },
        };
    };
    TaskManagerImpl.prototype.is_blocking = function (rbid) {
        return this.interpreter_n_blockers_lookup[rbid] != undefined;
    };
    TaskManagerImpl.prototype.dispose_task = function (task_id) {
        this.tasks_needing_dispose.push(task_id);
    };
    TaskManagerImpl.prototype.dispose_tasks_given = function (match) {
        var dispose_tasks = _filter(this.tasks, match);
        var dispose_ids = [];
        for (var i = 0; i < dispose_tasks.length; i++) {
            dispose_ids.push(dispose_tasks[i].id);
        }
        this.tasks_needing_dispose = this.tasks_needing_dispose.concat(dispose_ids);
    };
    TaskManagerImpl.prototype.now = function () {
        if (typeof performance !== 'undefined' && performance.now != undefined) {
            return performance.now();
        }
        return new Date().getTime();
    };
    TaskManagerImpl.prototype.deterministic_task_update = function (t) {
        var now = this.now();
        if (t.n_ticks == undefined || t.start_tick == undefined) {
            t.n_ticks = 0;
            t.start_tick = now;
            if (t.on_start) {
                t.on_start();
            }
            return;
        }
        var time_delta_ms = this.ms_per_update;
        if (t.lifetime == undefined) {
            if (t.on_tick) {
                t.on_tick(time_delta_ms);
            }
            return;
        }
        var ticks_to_live = t.lifetime / this.ms_per_update;
        var amount_done = t.n_ticks / ticks_to_live;
        if (amount_done >= 1) {
            if (t.on_end) {
                t.on_end(time_delta_ms);
            }
            this.dispose_task(t.id);
            return;
        }
        if (t.on_tick) {
            t.on_tick(time_delta_ms, amount_done);
        }
        t.n_ticks += 1;
    };
    TaskManagerImpl.prototype.wall_clock_task_update = function (t) {
        var now = this.now();
        if (t.start_tick == undefined || t.previous_tick == undefined) {
            t.start_tick = now;
            t.previous_tick = now;
            if (t.on_start) {
                t.on_start();
            }
            return;
        }
        var time_delta_ms = now - t.previous_tick;
        if (t.lifetime == undefined) {
            if (t.on_tick) {
                t.on_tick(time_delta_ms);
            }
            t.previous_tick = now;
            return;
        }
        var amount_done = (now - t.start_tick) / t.lifetime;
        if (t.lifetime <= now - t.start_tick) {
            if (t.on_end) {
                t.on_end(time_delta_ms);
            }
            this.dispose_task(t.id);
            return;
        }
        if (t.on_tick) {
            t.on_tick(time_delta_ms, amount_done);
        }
        t.previous_tick = now;
    };
    TaskManagerImpl = tslib_1.__decorate([
        inversify_1.injectable(),
        tslib_1.__param(0, inversify_1.inject(di_symbols_1.BINDING.EventBus)),
        tslib_1.__param(1, inversify_1.inject(di_symbols_1.BINDING.RuntimeData)),
        tslib_1.__param(2, inversify_1.inject(di_symbols_1.BINDING.Util)),
        tslib_1.__metadata("design:paramtypes", [Object, Object, Object])
    ], TaskManagerImpl);
    return TaskManagerImpl;
}());
exports.TaskManagerImpl = TaskManagerImpl;
